import numpy as np
from typing import List


def unity(x):
    return x


def quadratic(x: np.ndarray, params: List):
    return (x - params[0]) * (x - params[1])


def rectify(x):
    return np.maximum(x, 0)



import numpy as np


def sigm(x):
    return 1 / (1 + np.exp(-x))


def wc_drift(x, **kwargs):
    e = x[0]
    i = x[1]

    params = kwargs["params"]

    tau = params["tau"]
    alpha = params["alpha"]
    beta = params["beta"]
    w = params["w"]
    thresh = params["thresh"]

    e_dot_p = params["T_e"] * (
        -e + sigm(-beta["e"] * (e * w["ee"] - i * w["ei"] - thresh["e"]))
    )
    i_dot_p = params["T_i"] * (
        -i + sigm(-beta["i"] * (-i * w["ii"] + e * w["ie"] - thresh["i"]))
    )

    new_state = np.array([e_dot_p, i_dot_p])

    return new_state


def wc_input(x, **kwargs):
    e = x[0]
    i = x[1]

    params = kwargs["params"]

    tau = params["tau"]
    alpha = params["alpha"]
    beta = params["beta"]
    w = params["w"]
    thresh = params["thresh"]

    u = kwargs["u"]

    e_dot_p = (
        params["T_e"]
        * (-e + sigm(-beta["e"] * (e * w["ee"] - i * w["ei"] - thresh["e"])))
        + u
    )
    i_dot_p = params["T_i"] * (
        -i + sigm(-beta["i"] * (-i * w["ii"] + e * w["ie"] - thresh["i"]))
    )

    new_state = np.array([e_dot_p, i_dot_p])

    return new_state



import networkx as nx
import matplotlib.pyplot as plt


class connectivity:
    def __init__(self, N, structure=None, **kwargs):
        if callable(structure):
            self.G = structure(N, kwargs["proportion"])
        else:
            self.G = nx.generators.random_graphs.erdos_renyi_graph(
                N, kwargs["proportion"]
            )

    @property
    def D(self):
        return nx.linalg.graphmatrix.incidence_matrix(self.G).todense().T

    @property
    def spectrum(self):
        # L = nx.laplacian_matrix(self.G)
        # e = np.linalg.eigvals(L.A)

        return nx.laplacian_spectrum(self.G)

    #%% PLOTTING METHODS

    def plot_incidence(self):

        plt.figure()
        plt.imshow(self.D.T)
        plt.title("Incidence")

    def plot_spectrum(self):
        plt.figure()
        plt.plot(self.spectrum)
        plt.title("Spectrum of Laplacian")

        plt.suptitle(f"Number of components: {nx.number_connected_components(self.G)}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 17:21:34 2020

@author: virati
Barebones class for dynamical systems
"""

import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
import scipy.signal as sig
from autodyn.utils.functions import unity

from autodyn.core.integrators.runge_kutta import rk_integrator


class system:
    def __init__(self, f, D: int = 3, net_graph: nx.Graph = None):
        self.x = np.zeros((D, 1))
        self.D = D
        self.f = f

        self.gen_connectivity()
        self.post_step = unity

    def set_post_step(self, func: callable):
        self.post_step = func

    def gen_connectivity(self):
        nodes = [1, 2, 3, 4, 5, 6]
        G = nx.Graph()
        G.add_nodes_from(nodes)
        G.add_weighted_edges_from(
            [
                (1, 2, 1),
                (2, 3, 1),
                (1, 6, 1),
                (1, 3, 1),
                (3, 4, 1),
                (4, 5, 1),
                (4, 6, 1),
                (5, 6, 1),
            ]
        )

        # L = nx.to_numpy_matrix(G)
        L = nx.linalg.laplacianmatrix.laplacian_matrix(G).todense()
        self.G = G
        self.L = L

    def simulate(self, T, dt=0.01, rasterize=True, **kwargs):
        tvect = np.arange(0, T, dt)
        controlled = False

        x_state = np.random.normal(0, 1, (self.D, 1))
        if "keep_positive" in kwargs.keys():
            x_state = np.abs(x_state)

        x_raster = []

        if "stim" in kwargs.keys():
            controlled = True

        for tidx, time in enumerate(tvect):
            if controlled:
                x_new = rk_integrator(
                    self.f, x_state, dt=0.01, u=kwargs["stim"][tidx], **kwargs
                )
            else:
                x_new = rk_integrator(self.f, x_state, dt=0.01, **kwargs)

            x_state = self.post_step(x_new)

            if rasterize:
                x_raster.append(x_state)

        if x_raster:
            self.raster = np.array(x_raster).squeeze()

    def plot_raster(self):
        plt.plot(self.raster)
        plt.show()

    def plot_phase(self):
        if self.D == 3:
            fig = plt.figure()
            ax = fig.gca(projection="3d")
            ax.plot(self.raster[:, 0], self.raster[:, 1], self.raster[:, 2])
            plt.draw()
            plt.title("Phase Portrait")
        else:
            fig = plt.figure()
            plt.plot(self.raster)
            plt.title("Trajectories in Time")

    def plot_measure(self):
        plt.figure()
        plt.plot(self.H(self.raster))
        plt.title("Measured Trajectories in Time")

    def plot_polar(self):
        plt.figure()
        plt.plot(np.real(self.raster[:, 0] * np.exp(1j * self.raster[:, 1])))
        plt.title("Measured Trajectories in Time")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 14 15:46:03 2021

@author: virati
Animator class
"""

import numpy as np
import scipy.signal as sig
from scipy.integrate import odeint
import matplotlib.pyplot as plt
import sys
sys.path.append('/home/virati/Dropbox/projects/Research/Control/autoDyn/lib/')

from dynSys import dsys, rk_integrator
from dynLib import consensus

from matplotlib.animation import FuncAnimation
plt.style.use('seaborn-pastel')

def rkint(f,x0,tvec,params,dt):
    x_state = np.copy(x0)
    x_state_raster = []
    
    for tt in tvec:
        x_state_raster.append(np.copy(x_state))
        x_state += rk_integrator(f,params,x_state,dt=dt)
        
    return np.array(x_state_raster)


def rksim(**kwargs):
    for key,val in kwargs:
        print(key)

def oscillator(params,x):
    c = params['c']
    
    return -1*np.dot(c,-(x-5)*x)


#osc_drive = lambda x,t,c: -1*np.dot(c,-(x-5)*x)# + u

T = 1001
T_end=10
t = np.linspace(0,T_end,T)

x0 = np.array([1.,0.])

w=1
coupling = np.array([[0.1,w*2],[-2,0.5]])

params = {'c': coupling}
sol = rkint(oscillator,x0,t,params=params,dt=T_end/T)
#sol= odeint(osc_drive,x0,t,args=(coupling,))

plt.figure()
plt.plot(t,sol)
plt.show()


#%%
def init():
    line.set_data([],[])
    return line,

def animate(ii):
    line.set_data(t[:ii],sol[:ii])
    return line,

fig = plt.figure()
ax = plt.axes(xlim=(0,4), ylim=(-2,2))
line, = ax.plot([],[],lw=3)
anim = FuncAnimation(fig, animate, init_func=init,frames=200,interval=20,blit=True)

anim.save('test.gif',writer='imagemagick')


import numpy as np

# Dynamics Functions
def consensus(x, **kwargs):
    new_x = np.zeros_like(x)
    D = kwargs["D"]
    if "W" in kwargs.keys():
        W = kwargs["W"]
    else:
        W = np.eye(D.shape[1])

    new_x = -np.dot(np.dot(D, np.dot(W, D.T)), x)

    return new_x


def lorenz(x, **kwargs):
    new_x = np.zeros_like(x)
    sigma = kwargs["sigma"]
    rho = kwargs["rho"]
    beta = kwargs["beta"]

    new_x[0] = sigma * (x[1] - x[0])
    new_x[1] = x[0] * (rho - x[2]) - x[1]
    new_x[2] = x[0] * x[1] - beta * x[2]

    return new_x


def controlled_hopf(x, g, u, **kwargs):
    new_x = np.zeros_like(x)

    r = x[0]
    theta = x[1]

    c = kwargs["c"]
    w = kwargs["w"]

    new_x[0] = -(r**2) * (r - c) - g * u
    new_x[1] = w

    return new_x


def single_hopf(x, **kwargs):
    new_x = np.zeros_like(x)

    r = x[0]
    theta = x[1]

    c = kwargs["c"]
    w = kwargs["w"]

    new_x[0] = -(r**2) * (r - c)
    new_x[1] = w + 0 * theta

    return new_x



from setuptools import setup, find_packages

setup(
    name="autodyn",
    version="0.1.4",
    description="A module for differentiable dynamical systems",
    author="Vineet Tiruvadi",
    author_email="virati@gmail.com",
    package_dir={"": "src"},
    packages=find_packages("src"),
    install_requires=["wheel"],  # external packages as dependencies
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 11 17:57:21 2019

@author: virati
Main library for autoDyn methods
This file will contain the primary JAX related methods that analyse variables of the (dyn_sys) class
"""

import numpy as npo
import jax.numpy as np
from jax import grad, jit, vmap, jvp
from numpy import ndenumerate
import matplotlib.pyplot as plt
from jax import jit, jacfwd, jacrev

import operator
import networkx as nx

# Generate a class that wraps functions
class operable:
    def __init__(self, f):
        self.f = f
        
    def __call__(self, x):
        return self.f(x)

'''
Wrapper function for Lie derivative L_f H to a specific order
Returns exactly the output lie derivative
'''
class liefn:
    def __init__(self,f,H,order):
        self.f = f
        self.H = H
        self.order = order
        
        c = [H]
        for ii in range(order):
            c.append(npo.dot(operable(jit(jacfwd(c[ii],argnums=0))),f))
        
        self.LfH = c[-1]
    
    def __call__(self,x):
        return self.LfH(x0).squeeze().sum(axis=npo.arange(self.order)+1).T
    
class FeedBack:
    def __init__(self,g,f,H,order):
        self.f = f
        self.g = g
        self.H = H
        self.order = order
        
        Lfn = [H]
        for ii in range(order):
            c.append(npo.dot(operable(jit(jacfwd(c[ii],argnums=0))),f))
        
        self.Lfn = c
    
    def check_zeros(self):
        # go up to order and check if Lg Lf H is zero at each stage
        for ii in self.order:
            liefn(self.g,self.Lfn[ii])
    
        
# Convert oeprators to the corresponding function
def op_to_function_op(op):
    def function_op(self, operand):
        def f(x):
            return op(self(x), operand(x))
        return operable(f)
    return function_op
 
for name, op in [(name, getattr(operator, name)) for name in dir(operator) if "__" in name]:
    try:
        op(1,2)
    except TypeError:
        pass
    else:
        setattr(operable, name, op_to_function_op(op))

#%% Lie Derivatives Block

def L_f_x(f,h,x):
    #return jacfwd(h)(x).squeeze()
    #print()
    #print(f(x))
    return npo.dot(jacfwd(h,argnums=0)(x).squeeze(),f(x))
    #return npo.dot(operable(vmap(jacfwd(h))),f)

def L_f(f,h):
    return npo.dot(operable(jacfwd(h,argnums=0)),operable(f))

def L_f_o(f,h,order=1):
    c = [h]
    for ii in range(order):
        c.append(npo.dot(operable(jacfwd(c[ii],argnums=0)),f))
    
    return c[-1]

def dotL_f(f,h,order=1):
    return npo.sum(L_d(h,f,order=order))

def brack_f_g(f,g):
    c = operable(jcb(f)) * g
    cinv = operable(jcb(g)) * f
    
    return cinv
    #print(c(np.array([1.,1.,1.])))
    #print(cinv(np.array([1.,1.,1.])))

#%% Drift function libraries
@operable
def f1(x):
    #return np.array([-x[1],-x[0],-x[2] + x[1]])
    return np.array([-x[1]**2 + x[2],-x[0]**3,-x[2]**2 + x[1]])

@operable
def f2(x):
    return np.array([-x[0] + x[1],x[1],x[2] - x[0]])

@operable
def s1(x,L):
    return -np.dot(L,x)

def g1(x):
    return x

def h(x):
    return 2*x[0] + 3*x[2]


#%%
# Do some testing of the above library
if __name__ == '__main__':
    x0 = np.array([1.,1.,2.]).reshape(-1,1)
    readout = liefn(f1,f2,order=3)
    print('Readout',readout(x0))

if __name__ == '__net__':
    s1_grad = jacfwd(s1,argnums=0)
    x0 = np.array([1.,1.,2.]).reshape(-1,1)
    L = npo.random.normal(0,1,size=(3,3))
    print('S1 Gradient',s1_grad(x0,L).squeeze())
    
if __name__ == '__tests__':
    f_grad = jacfwd(f1,argnums=0)
    x0 = np.array([1.,1.,2.]).reshape(-1,1)
    print('F1, only\n',f_grad(x0).squeeze())
    #SUCCESS!
    
    f_all_grad = jacfwd(f1 + f2,argnums=0)
    print('F1+F2\n',f_all_grad(x0).squeeze())
    #SUCCESS!
    
    print('Lie deriv (actual)\n',L_f_x(f1,f2,x0))
    #SUCCESS!
    
    #lie
    do_order = 3
    L_f1_f2 = L_f_o(f1,f2,order=do_order)
    #print('Lie (jax)\n',np.sum(L_f1_f2(x0).squeeze(),axis=1).T) #if we are only doing 1st order Lie derivative
    #print('Lie (jax)\n',np.sum(L_f1_f2(x0).squeeze(),axis=(1,2)).T) #if order is 2
    #print('Lie (jax) DYNSUM\n',np.sum(L_f1_f2(x0).squeeze(),axis=npo.arange(do_order)+1).T) #if order is 2
    print('Lie (jax) DYNSUM\n',L_f1_f2(x0).squeeze().sum(axis=npo.arange(do_order)+1).T) #if order is 2
    #SUCCESS!!!!
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 22 19:53:34 2020

@author: virati
"""

import networkx as nx
from mayavi import mlab
import numpy as np
from ..base.dynSys import dsys

def plain_render_graph(H):
    # reorder nodes from 0,len(G)-1
    G=nx.convert_node_labels_to_integers(H)
    # 3d spring layout
    pos=nx.spring_layout(G,dim=3)
    # numpy array of x,y,z positions in sorted node order
    xyz=np.array([pos[v] for v in sorted(G)])
    # scalar colors
    scalars=np.array(G.nodes())+5
    
    mlab.figure(1, bgcolor=(0, 0, 0))
    mlab.clf()
    
    pts = mlab.points3d(xyz[:,0], xyz[:,1], xyz[:,2],
                        scalars,
                        scale_factor=0.1,
                        scale_mode='none',
                        colormap='Blues',
                        resolution=20)
    
    
    pts.mlab_source.dataset.lines = np.array(G.edges())
    tube = mlab.pipeline.tube(pts, tube_radius=0.01)
    mlab.pipeline.surface(tube, color=(0.8, 0.8, 0.8),opacity=0.1)
    
def render_graph(H,read,write):
    # reorder nodes from 0,len(G)-1
    G=nx.convert_node_labels_to_integers(H)
    # 3d spring layout
    pos=nx.spring_layout(G,dim=3)
    # numpy array of x,y,z positions in sorted node order
    xyz=np.array([pos[v] for v in sorted(G)])
    # scalar colors
    scalars=np.array(G.nodes())+5
    
    mlab.figure(1, bgcolor=(0, 0, 0))
    mlab.clf()
    
    pts = mlab.points3d(xyz[:,0], xyz[:,1], xyz[:,2],
                        scalars,
                        scale_factor=0.1,
                        scale_mode='none',
                        colormap='Blues',
                        resolution=20)
    pts_reads = mlab.points3d(xyz[read,0], xyz[read,1], xyz[read,2],
                    scalars[read],
                    scale_factor=0.1,
                    scale_mode='none',
                    color=(0.0,1.0,0.0),
                    resolution=20)
    pts_reads = mlab.points3d(xyz[write,0], xyz[write,1], xyz[write,2],
                scalars[write],
                scale_factor=0.1,
                scale_mode='none',
                color=(1.0,0.0,0.0),
                resolution=20)
    
    
    pts.mlab_source.dataset.lines = np.array(G.edges())
    tube = mlab.pipeline.tube(pts, tube_radius=0.01)
    mlab.pipeline.surface(tube, color=(0.8, 0.8, 0.8),opacity=0.1)
    
    mlab.savefig('mayavi2_spring.png')
    #mlab.show() # interactive window



def render_system(system : dsys):
    pass


import numpy as np
from typing import Optional

# Integrator
def rk_integrator(
    f_dyn: callable,
    state: np.ndarray,
    u: Optional[np.ndarray] = None,
    dt: float = 0.01,
    **kwargs,
) -> np.ndarray:

    """
    Basic Runge-Kutta integrator with dynamics, state, and control parameters

    Inputs
    ------
        f_dyn : Callable
            The drift dynamics as a function

        state : np.ndarray
            The state vector

        u : Union[int, np.ndarray]
            The control signal

        dt

    Outputs
    -------
        new_state : np.ndarray
            The new state after integrating through the effects of the dynamics

    """
    if u is None:
        u = 0

    k1 = f_dyn(state, u=u, **kwargs) * dt
    k2 = f_dyn(state + 0.5 * k1, u=u, **kwargs) * dt
    k3 = f_dyn(state + 0.5 * k2, u=u, **kwargs) * dt
    k4 = f_dyn(state + k3, u=u, **kwargs) * dt

    new_state = state + (k1 + 2 * k2 + 2 * k3 + k4) / 6

    return new_state
